"""
Resizable thread pool.

"""
__author__ = "Andrew Gillis"
__license__ = "http://www.opensource.org/licenses/mit-license.php"
__maintainer__ = "Andrew Gillis"
__email__ = "gillis.andrewj@gmail.com"

import threading
try:
    import queue
except ImportError:
    import Queue as queue

from asyncthreads import completion


class ThreadPool(object):

    """
    Resizable thread pool class.

    Maintain a pool of threads that execute tasks given to submit().  The
    number of worker threads can be changed by calling resize_pool() to set the
    number of workers explicitly, or by setting a max_size greater than
    min_size and letting to pool automatically change its size.

    """

    def __init__(self, min_size, max_size=None, queue_size=0):
        """Initialize the thread pool and start the minimum number of threads.

        If max_size is specified and is larger than min_size, then the pool
        will automatically resize.

        Arguments:
        min_size    -- Minimum number of threads to have in pool.
        max_size    -- Optional.  Maximum size pool can grow to.
        queue_size  -- Maximum items allowed in task queue.   If queue_size is
                       less than or equal to zero, the queue size is infinite.

        """
        assert(isinstance(min_size, int)), 'min_size must be int'
        assert(isinstance(max_size, int)), 'max_size must be int'

        self._shutdown = False
        self._threads = 0
        self._task_queue = queue.Queue(queue_size)
        self._resize_lock = threading.Lock()
        self._idle_counter = ThreadCounter()

        self._min_size = min_size
        if max_size and max_size > min_size:
            # Pool is automatically resizeable.
            self._max_size = max_size
        else:
            # Pool is manually resizeable only.
            self._max_size = min_size

    def submit(self, function, args=None, callback=None, result_q=None):
        """Insert a task function into the queue.

        Arguments:
        function -- Function to execute.
        args     -- Argument tuple to pass to function.
        callback -- Optional callback that takes return from func as argument.
        result_q -- Optional response queue to place Completion on.

        Return:
        Completion object to wait on and retrieve result of function.

        """
        future = completion.Completion()
        self.submit_task(
            completion._Task(future, function, args, callback, result_q))
        return future

    def submit_task(self, task):
        """Insert a Task object into the queue.

        Useful for submitting task when Task object already exists.

        """
        with self._resize_lock:
            if self._shutdown:
                raise RuntimeError('cannot submit tasks after shutdown')

            self._task_queue.put(task, False)
            self._adjust_thread_count()

    def threads(self):
        """Return the number of threads in the pool."""
        return self._threads

    def idle_threads(self):
        """Return the current number of idle threads."""
        return int(self._idle_counter)

    def qsize(self):
        """Return the number of items currently in task queue."""
        return self._task_queue.qsize()

    def shutdown(self, wait_for_threads=True, drop_queued_tasks=False):
        """Shutdown the thread pool.

        Optionally wait for worker threads to exit, and optionally drop any
        queued tasks.

        Arguments:
        wait_for_threads  -- Wait for all threads to complete.
        drop_queued_tasks -- Discard any tasks that are on the task queue, but
                             have not yet been processed.

        Return:
        True if exited successfully, False if shutdown was already called.

        """
        with self._resize_lock:
            if self._shutdown:
                return False

            self._shutdown = True

            while drop_queued_tasks:
                # Discard remaining tasks on queue.
                try:
                    self._task_queue.get(False)
                    self._task_queue.task_done()
                except queue.Empty:
                    break

            # Enqueue a shutdown request for every thread.
            while self._threads:
                self._threads -= 1
                self._task_queue.put(None)

            # Wait until all shutdown requests are processed.
            if wait_for_threads:
                self._task_queue.join()

        return True

    def _adjust_thread_count(self):
        """If more items in the queue than idle threads, add thread.

        """
        idle = int(self._idle_counter)
        qsize = self._task_queue.qsize()

        # Grow pool if there are more items in queue than there idle threads.
        if qsize > idle:
            if self._threads < self._max_size:
                t = threading.Thread(target=_worker,
                                     args=(self._task_queue,
                                           self._idle_counter))
                t.daemon = True
                t.start()
                self._threads += 1

        # Shrink pool if fewer items in queue than idle threads above minimum
        # size, and more idle threads than the minimum pool size.
        elif idle > self._min_size and qsize < idle:
            self._task_queue.put(None)
            self._threads -= 1

        return self._threads

    def resize_pool(self, new_size):
        """Resize the thread pool to have the specified number of threads.

        """
        if new_size < 0:
            raise ValueError('pool size must be >= 0')

        with self._resize_lock:
            if new_size > self._threads:
                # Growing pool.
                args = (self._task_queue, self._idle_counter)
                while self._threads < new_size:
                    self._threads += 1
                    t = threading.Thread(target=_worker, args=args)
                    t.daemon = True
                    t.start()
            else:
                # Shrinking pool.
                while self._threads > new_size:
                    self._threads -= 1
                    self._task_queue.put(None)


class ThreadCounter(object):

    def __init__(self):
        self._lock = threading.Lock()
        self._count = 0

    def __int__(self):
        return self._count

    def incr(self):
        with self._lock:
            self._count += 1

    def decr(self):
        with self._lock:
            self._count -= 1

    def set(self, value):
        with self._lock:
            self._count = value


def _worker(task_queue, idle_counter):
    """Retrieve the next task and execute it, calling the callback if any.

    """
    while True:
        try:
            task = task_queue.get(False)
        except queue.Empty:
            idle_counter.incr()
            # Wait forever.
            task = task_queue.get()
            idle_counter.decr()

        # Exit if told to quit.
        if task is None:
            task_queue.task_done()
            break

        # Execute task.
        task.execute()

        del task
        task_queue.task_done()
